"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2018 Fernando Serena.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import re

from agora import Wrapper
from agora.engine.plan.agp import extend_uri
from agora_wot.blocks.td import TD

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.gql.schema')


def name(prefixed_name):
    return prefixed_name.split(':')[1]


def transform_term(t):
    return t if t.isupper() else t.title()


def convert(name):
    s1 = re.sub('(.)([A-Z][a-z]*)', r'\1_\2', name)
    return ''.join(map(lambda x: transform_term(x), s1.split('_')))


def title(prefixed_name):
    return convert(name(prefixed_name))


def attr_type(fountain, p, all_type_names):
    p_dict = fountain.get_property(p)
    p_range = p_dict['range']
    if p_dict['type'] == 'data':
        value_type = p_range[0]
        if value_type.startswith('xsd'):
            value_type = title(value_type)
            if value_type == 'Integer':
                value_type = 'Int'
        else:
            value_type = 'String'
        return value_type

    super_range = filter(lambda x: not set.intersection(set(fountain.get_type(x)['super']), set(p_range)), p_range)
    if super_range:
        if len(super_range) > 1:
            range_type_name = 'Union' + '_'.join(map(lambda x: title(x), sorted(super_range)))
            return '[{}]'.format(range_type_name)
        else:
            concrete_range = filter(lambda x: not set.intersection(set(fountain.get_type(x)['sub']), set(p_range)),
                                    p_range)

            range_type_name = super_range[0]
            gql_type = all_type_names[range_type_name]
            if len(concrete_range) > 1:
                interface = 'I' + gql_type
                gql_type = interface

            return '[{}]'.format(gql_type)


def get_abstract_types(fountain):
    properties = fountain.properties
    for p in properties:
        p_dict = fountain.get_property(p)
        p_range = p_dict['range']
        if p_dict['type'] == 'object':
            super_range = filter(lambda x: not set.intersection(set(fountain.get_type(x)['super']), set(p_range)),
                                 p_range)
            if super_range:
                if len(super_range) > 1:
                    yield (p, {'type': 'Union', 'of': super_range})
                else:
                    concrete_range = filter(
                        lambda x: not set.intersection(set(fountain.get_type(x)['sub']), set(p_range)),
                        p_range)

                    if len(concrete_range) > 1:
                        yield (
                            p, {'type': 'Interface', 'for': concrete_range, 'base': super_range[0]})


def serialize_type(fountain, t, all_type_names, abstract_types):
    t_dict = fountain.get_type(t)
    t_props = t_dict['properties']

    abstract_refs = set.intersection(set(abstract_types.keys()), set(t_dict['refs']))
    implements = [abstract_types[p]['base'] for p in abstract_refs if
                  abstract_types[p]['type'] == 'Interface']

    attr_lines = ['\t{}: {}'.format(name(p), attr_type(fountain, p, all_type_names)) for p in t_props]
    attr_lines_str = '\n'.join(attr_lines)

    res = ''

    if t in implements:
        res += 'interface {} '.format('I' + all_type_names[t])
        res += '{\n%s\n}\n' % attr_lines_str

    res += 'type {} '.format(all_type_names[t])
    if implements:
        implements = ['I' + all_type_names[i] for i in implements]
        res += 'implements ' + ', '.join(implements) + ' '

    res += '{\n%s\n}' % attr_lines_str
    return res


def query_args(args):
    args = ', '.join(['{}: String'.format(arg.lstrip('$')) for arg in args])
    return '({})'.format(args) if args else ''


def serialize_queries(type_args):
    query_lines = ['\t{}{}: [{}]'.format(t, query_args(args), t) for t, args in type_args.items()]
    return 'type Query {\n%s\n}' % '\n'.join(query_lines)


def create_gql_schema(gateway):
    log.info('Building GraphQL schema from Agora...')
    fountain = Wrapper(gateway.agora.fountain)

    types = filter(lambda x: fountain.get_type(x)['properties'], sorted(fountain.types))
    all_type_names = {}
    for t in types:
        t_title = title(t)
        if t_title in all_type_names:
            t_title = ''.join(map(lambda x: x.title(), t.split(':')))

        all_type_names[t] = t_title
        all_type_names[t_title] = t

    t_params = {}
    prefixes = gateway.agora.fountain.prefixes
    for t in types:
        t_uri = extend_uri(t, prefixes)
        t_ted = gateway.discover("""SELECT * WHERE { [] a <%s>}""" % t_uri, strict=True, lazy=False)
        if t_ted.ecosystem.roots:
            params = set()
            for root in t_ted.ecosystem.roots:
                if isinstance(root, TD):
                    root_vars = t_ted.ecosystem.root_vars(root)
                    td_vars = filter(lambda x: x != '$item' and x != '$parent', root_vars)
                    params.update(set(td_vars))

            t_params[all_type_names[t]] = params

    abstract_types = dict(get_abstract_types(fountain))
    for abstract in abstract_types.values():
        if abstract['type'] == 'Union':
            union_type_names = map(lambda x: title(x), sorted(abstract['of']))
            union_name = 'Union' + '_'.join(union_type_names)
            all_type_names[union_name] = union_type_names

    types_str = '\n'.join(
        filter(lambda x: x, [serialize_type(fountain, t, all_type_names, abstract_types) for t in types]))
    query_str = serialize_queries(t_params)
    schema_str = "schema {\n\tquery: Query\n}"

    unions_dict = {k: v for k, v in all_type_names.items() if k.startswith('Union')}
    unions = ['union {}= {}'.format(u, '|'.join(union_types)) for u, union_types in unions_dict.items()]
    unions_str = '\n'.join(unions)

    res = '\n'.join([types_str, unions_str, query_str, schema_str])

    return res



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2018 Fernando Serena.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
from expiringdict import ExpiringDict
from graphql import parse, build_ast_schema, MiddlewareManager, Source, validate, execute
from graphql.execution import ExecutionResult

from agora_graphql.gql.executor import AgoraExecutor
from agora_graphql.gql.middleware import AgoraMiddleware
from agora_graphql.gql.schema import create_gql_schema

__author__ = 'Fernando Serena'


class GraphQLProcessor(object):
    def __init__(self, gateway, schema_path=None, data_gw_cache=None, **kwargs):
        self.__gateway = gateway

        if schema_path:
            with open(schema_path) as f:
                source = f.read()
        else:
            self.__schema_source = create_gql_schema(gateway)
            source = self.__schema_source

        document = parse(source)
        self.__schema = build_ast_schema(document)
        abstract_types = filter(lambda x: hasattr(x, 'resolve_type'), dict(self.__schema.get_type_map()).values())
        for at in abstract_types:
            at.resolve_type = self.__resolve_type

        self.__executor = AgoraExecutor(gateway)

        if not data_gw_cache:
            data_gw_cache = {'max_age_seconds': 300, 'max_len': 1000000}

        self.expiring_dict = ExpiringDict(**data_gw_cache)
        middleware = AgoraMiddleware(gateway, data_gw_cache=self.expiring_dict, **kwargs)
        self.__middleware = MiddlewareManager(middleware)

    def __resolve_type(self, *args, **kwargs):
        m = self.middleware.middlewares[0]
        return m.resolve_type(*args, **kwargs)

    @property
    def schema_text(self):
        return self.__schema_source

    @property
    def middleware(self):
        return self.__middleware

    @property
    def executor(self):
        return self.__executor

    @property
    def middleware(self):
        return self.__middleware

    @property
    def schema(self):
        return self.__schema

    def query(self, q):
        try:
            source = Source(q, name='GraphQL request')
            ast = parse(source)
            validation_errors = validate(self.schema, ast)
            if validation_errors:
                return ExecutionResult(
                    errors=validation_errors,
                    invalid=True,
                )
        except Exception as e:
            return ExecutionResult(errors=[e], invalid=True)

        try:
            return execute(self.__schema,
                           ast,
                           root_value=None,
                           variable_values={},
                           operation_name=None,
                           context_value={
                               'query': q,
                               'introspection': 'introspection' in q.lower()
                           },
                           middleware=self.__middleware,
                           executor=self.__executor
                           )
        except Exception as e:
            return ExecutionResult(errors=[e], invalid=True)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2018 Fernando Serena.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

from agora_gw import Gateway
from flask import Flask
from flask_cors import CORS

from agora_graphql.gql import GraphQLProcessor
from agora_graphql.server.view import AgoraGraphQLView

__author__ = 'Fernando Serena'


def build(**kwargs):
    app = Flask(__name__)
    CORS(app)

    schema_path = kwargs.get('schema', {}).get('path', None)

    gw = Gateway(**kwargs['gateway'])
    gql_processor = GraphQLProcessor(gw, schema_path, data_gw_cache=kwargs.get('gw_cache', None))

    app.add_url_rule('/graphql',
                     view_func=AgoraGraphQLView.as_view('graphql', schema=gql_processor.schema,
                                                        executor=gql_processor.executor,
                                                        middleware=gql_processor.middleware,
                                                        graphiql=True))

    return app



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2018 Fernando Serena.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
from agora.engine.plan import AGP
from graphql import parse
from graphql.language.ast import Field
from shortuuid import uuid

from agora_graphql.misc import match

__author__ = 'Fernando Serena'


def identify_property(parent_type, p, fountain):
    parent_props = fountain.get_type(parent_type)['properties']
    try:
        return match(p, parent_props).pop()
    except IndexError:
        pass


def identify_parent_type(selection, fountain):
    cand_parent_types = list(match(selection, fountain.types))
    field_names = [f.name.value for f in selection.selection_set.selections if isinstance(f, Field)]

    for t in cand_parent_types:
        if all([identify_property(t, p, fountain) for p in field_names]):
            return t


def process_selection(fountain, selection, parent_var=None, parent_type=None, parent_prop=None, root_mode=True):
    var = None
    ty = None
    prop = None

    try:
        if not parent_type:
            ty = identify_parent_type(selection, fountain)
            if ty:
                if not parent_var:
                    var = '?v' + str(uuid())
                else:
                    var = parent_var
                tp = '{} {} {}'.format(var, 'a', ty)
                yield tp

            if not root_mode and not parent_prop:
                prop = list(match(selection, fountain.properties)).pop()
                if prop:
                    if not parent_var:
                        parent_var = '?v' + str(uuid())
                    var = '?v' + str(uuid())
                    tp = '{} {} {}'.format(parent_var, prop, var)
                    yield tp

        if parent_type or parent_prop:
            if parent_type:
                prop = list(match(selection, fountain.get_type(parent_type)['properties'])).pop()
            else:
                parent_range = fountain.get_property(parent_prop)['range']
                cand_props = reduce(lambda x, y: x.union(set(fountain.get_type(y)['properties'])), parent_range, set())
                prop = list(match(selection, cand_props)).pop()
            if prop:
                var = '?v' + str(uuid())
                tp = '{} {} {}'.format(parent_var, prop, var)
                yield tp

        if not root_mode and selection.selection_set:
            for s in selection.selection_set.selections:
                for tp in process_selection(fountain, s, parent_var=var, parent_type=ty, parent_prop=prop):
                    yield tp
    except IndexError:
        pass


def sparql_from_graphql(fountain, gql_query, root_mode=False):
    ast = parse(gql_query)
    tps = []
    for d in ast.definitions:
        if d.operation == 'query':
            var = '?v' + str(uuid())
            for s in d.selection_set.selections:
                tps.extend(list(process_selection(fountain, s, parent_var=var, root_mode=root_mode)))

    agp = AGP(set(tps), prefixes=fountain.prefixes)
    query = 'SELECT DISTINCT * WHERE {}'.format(agp)
    return query
